"""
Trust Region Subproblem Solver
====

Specifically, the call
    d, gnew, crvmin = trsbox(xopt, g, hess, sl, su, delta)
produces a new vector d which (approximately) solves the trust region subproblem:
    min_{d}  g'*d + 0.5*d'*hess*d
    s.t.    ||d|| <= delta
            sl <= xopt + d <= su
The other outputs: gnew is the gradient of the model at d, and crvmin has
information about the curvature of the model at the solution.

We also provide a function for maximising the absolute value of a linear function
inside a similar trust region - this is useful for geometry steps.
The call
    x = trsbox_geometry(xbase, c, g, lower, upper, delta)
solves
    min_x  abs(c + g' * (x - xbase))
    s.t.  lower <= x <= upper
          ||x-xbase|| <= Delta
With this value, the variable d=x-xbase solves the problem
    min_s  abs(c + g' * d)
    s.t.   lower <= xbase + d <= upper
          ||d|| <= delta

Notes
----
The solver trsbox is an implementation of the routine TRSBOX from BOBYQA (Powell, 2009).
Some modifications to the termination conditions are from the equivalent routine
from DFBOLS (Zhang et al, 2010).


This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.

The development of this software was sponsored by NAG Ltd. (http://www.nag.co.uk)
and the EPSRC Centre For Doctoral Training in Industrially Focused Mathematical
Modelling (EP/L015803/1) at the University of Oxford. Please contact NAG for
alternative licensing.

"""

# Ensure compatibility with Python 2
from __future__ import absolute_import, division, print_function, unicode_literals

import logging
from math import sqrt
import numpy as np
try:
    import trustregion
    USE_FORTRAN = True
except ImportError:
    # Fall back to Python implementation
    USE_FORTRAN = False


from .util import sumsq


__all__ = ['trsbox', 'trsbox_geometry']

ZERO_THRESH = 1e-14


def trsbox(xopt, g, hess, sl, su, delta, use_fortran=USE_FORTRAN):
    if use_fortran:
        return trustregion.solve(g, hess, delta,
                                 sl=np.minimum(sl - xopt, -ZERO_THRESH),
                                 su=np.maximum(su - xopt, ZERO_THRESH),
                                 verbose_output=True)

    n = xopt.size
    assert xopt.shape == (n,), "xopt has wrong shape (should be vector)"
    assert g.shape == (n,), "g and xopt have incompatible sizes"
    assert hess.shape == (n,n), "hess and xopt have incompatible sizes"
    assert sl.shape == (n,), "sl and xopt have incompatible sizes"
    assert su.shape == (n,), "su and xopt have incompatible sizes"
    assert delta > 0.0, "delta must be strictly positive"
    # Assume g and hess have full quadratic model for objective
    # i.e. skip straight to label 8 in DFBOLS version

    # The sign of G(I) gives the sign of the change to the I-th variable
    # that will reduce Q from its value at XOPT. Thus XBDI(I) shows whether
    # or not to fix the I-th variable at one of its bounds initially, with
    # NACT being set to the number of fixed variables. D and GNEW are also
    # set for the first iteration. DELSQ is the upper bound on the sum of
    # squares of the free variables. QRED is the reduction in Q so far.

    iterc = 0
    nact = 0  # number of fixed variables

    xbdi = np.zeros((n,), dtype=int)  # fix x_i at bounds? [values -1, 0, 1]
    xbdi[(xopt <= sl) & (g >= 0.0)] = -1
    xbdi[(xopt >= su) & (g <= 0.0)] = 1

    d = np.zeros((n,))
    s = np.zeros((n,))
    gnew = g.copy()
    qred = 0.0
    delsq = delta ** 2
    crvmin = -1.0
    beta = 0.0  # label 20

    need_alt_trust_step = False  # will either quit main CG loop to finish, or do alternative step
    MAX_LOOP_ITERS = 100 * n ** 2  # avoid infinite loops
    # while True:  # main CG loop [label 30]
    for ii in range(MAX_LOOP_ITERS):
        s[xbdi != 0] = 0.0
        if beta == 0.0:
            s[xbdi == 0] = -gnew[xbdi == 0]
        else:
            s[xbdi == 0] = beta * s[xbdi == 0] - gnew[xbdi == 0]
        stepsq = sumsq(s)

        if stepsq == 0.0:
            need_alt_trust_step = False
            break  # break and quit

        if beta == 0.0:
            gredsq = stepsq
            itermax = iterc + n - nact

        if iterc == 0:
            gredsq0 = gredsq

        # Exit conditions
        if gredsq <= min(1.0e-6 * gredsq0, 1.0e-18) or gredsq * delsq <= min(1.0e-6 * qred ** 2, 1.0e-18):  # DFBOLS
            need_alt_trust_step = False
            break  # break and quit

        # Multiply the search direction by the second derivative matrix of Q and
        # calculate some scalars for the choice of steplength. Then set BLEN to
        # the length of the the step to the trust region boundary and STPLEN to
        # the steplength, ignoring the simple bounds.

        hs = hess.dot(s)

        # label 50
        ds = np.dot(s[xbdi == 0], d[xbdi == 0])
        shs = np.dot(s[xbdi == 0], hs[xbdi == 0])
        resid = delsq - sumsq(d[xbdi == 0])
        if resid <= 0.0:
            need_alt_trust_step = True
            break  # break and calculate alt step instead

        temp = sqrt(stepsq * resid + ds ** 2)
        blen = (resid / (temp + ds) if ds >= 0.0 else (temp - ds) / stepsq)
        stplen = (blen if shs <= 0.0 else min(blen, gredsq / shs))

        # Exit condition
        if stplen <= 1.0e-30:  # DFBOLS
            need_alt_trust_step = False
            break  # break and quit

        # Reduce STPLEN if necessary in order to preserve the simple bounds,
        # letting IACT be the index of the new constrained variable.
        iact = None
        for i in range(n):
            if s[i] != 0.0:
                temp = (su[i] - xopt[i] - d[i] if s[i] > 0.0 else sl[i] - xopt[i] - d[i]) / s[i]
                if temp < stplen:
                    stplen = temp
                    iact = i

        # Update CRVMIN, GNEW and D. Set SDEC to the decrease that occurs in Q.
        sdec = 0.0
        if stplen > 0.0:
            iterc += 1
            temp = shs / stepsq
            if iact is None and temp > 0.0:
                crvmin = min(crvmin, temp) if crvmin != -1.0 else temp
            ggsav = gredsq
            gnew += stplen * hs
            d += stplen * s
            gredsq = sumsq(gnew[xbdi == 0])
            sdec = max(stplen * (ggsav - 0.5 * stplen * shs), 0.0)
            qred += sdec

        # Restart the conjugate gradient method if it has hit a new bound.
        if iact is not None:
            nact += 1
            xbdi[iact] = (1 if s[iact] >= 0.0 else -1)
            delsq = delsq - d[iact] ** 2
            if delsq <= 0.0:
                need_alt_trust_step = True
                break  # break and calculate alt step instead
            beta = 0.0  # label 20
            continue  # restart loop (new CG iteration)

        # If STPLEN is less than BLEN, then either apply another conjugate
        # gradient iteration or RETURN.
        if stplen >= blen:
            need_alt_trust_step = True
            break  # break and calculate alt step instead

        # Exit condition
        if iterc == itermax or sdec <= 1.0e-6 * qred:  # DFBOLS
            need_alt_trust_step = False
            break  # break and quit

        beta = gredsq / ggsav
        continue  # new CG iteration
    # end of CG loop

    # either done or need to take and alternative step
    if need_alt_trust_step:
        crvmin = 0.0
        d, gnew = alt_trust_step(n, xopt, hess, sl, su, d, xbdi, nact, gnew, qred)
        return d, gnew, crvmin
    else:
        return d_within_bounds(d, xopt, sl, su, xbdi), gnew, crvmin


# Alternative Trust Region Step (label 100 of TRSBOX in BOBYQA, where crvmin=0)
def alt_trust_step(n, xopt, hess, sl, su, d, xbdi, nact, gnew, qred):
    MAX_LOOP_ITERS = 100 * n ** 2  # avoid infinite loops
    # while True:  # label 100 here
    for ii in range(MAX_LOOP_ITERS):
        if nact >= n - 1:
            return d_within_bounds(d, xopt, sl, su, xbdi), gnew

        # Prepare for the alternative iteration by calculating some scalars
        # and by multiplying the reduced D by the second derivative matrix of
        # Q, where S holds the reduced D in the call of GGMULT.
        s = np.zeros((n,))
        s[xbdi == 0] = d[xbdi == 0]
        dredsq = sumsq(d[xbdi == 0])
        dredg = np.dot(d[xbdi == 0], gnew[xbdi == 0])
        gredsq = sumsq(gnew[xbdi == 0])

        # Label 210 (crvmin = 0, itcsav = iterc)
        hs = hess.dot(s)

        hred = hs.copy()
        # quit 210 by goto 120

        # Let the search direction S be a linear combination of the reduced D
        # and the reduced G that is orthogonal to the reduced D.
        restart_alt_loop = False  # once the below loop finishes, quit unless need to go again
        # while True:  # label 120
        for jj in range(MAX_LOOP_ITERS):
            temp = gredsq * dredsq - dredg ** 2
            if temp <= 1.0e-4 * qred ** 2:
                restart_alt_loop = False
                break  # quit inner label 120 loop and return results
            temp = sqrt(temp)
            s = np.zeros((n,))
            s[xbdi == 0] = (dredg * d[xbdi == 0] - dredsq * gnew[xbdi == 0]) / temp
            sredg = -temp

            # By considering the simple bounds on the variables, calculate an upper
            # bound on the tangent of half the angle of the alternative iteration,
            # namely ANGBD, except that, if already a free variable has reached a
            # bound, there is a branch back to label 100 after fixing that variable.
            free_variable_reached_bound = False
            angbd = 1.0
            iact = None
            for i in range(n):
                if xbdi[i] == 0:
                    tempa = xopt[i] + d[i] - sl[i]
                    tempb = su[i] - xopt[i] - d[i]
                    if tempa <= 0.0:
                        nact += 1
                        xbdi[i] = -1
                        free_variable_reached_bound = True
                        break  # skip the rest of this for loop
                    elif tempb <= 0.0:
                        nact += 1
                        xbdi[i] = 1
                        free_variable_reached_bound = True
                        break  # skip the rest of this for loop
                    ssq = d[i] ** 2 + s[i] ** 2
                    temp = ssq - (xopt[i] - sl[i]) ** 2
                    if temp > 0.0:
                        temp = sqrt(temp) - s[i]
                        if angbd * temp > tempa:
                            angbd = tempa / temp
                            iact = i
                            xsav = -1
                    temp = ssq - (su[i] - xopt[i]) ** 2
                    if temp > 0.0:
                        temp = sqrt(temp) + s[i]
                        if angbd * temp > tempb:
                            angbd = tempb / temp
                            iact = i
                            xsav = 1
            # End for loop
            if free_variable_reached_bound:  # deal with break conditions above
                restart_alt_loop = True
                break  # quit inner label 120 loop and restart alt iteration loop (label 100)

            # Label 210 (crvmin = 0, itcsav < iterc since iterc+=1 earlier)
            hs = hess.dot(s)

            # Label 150
            # Calculate HHD and some curvatures for the alternative iteration.
            shs = np.sum(s[xbdi == 0] * hs[xbdi == 0])
            dhs = np.sum(d[xbdi == 0] * hs[xbdi == 0])
            dhd = np.sum(d[xbdi == 0] * hred[xbdi == 0])

            # Seek the greatest reduction in Q for a range of equally spaced values
            # of ANGT in [0,ANGBD], where ANGT is the tangent of half the angle of
            # the alternative iteration.
            redmax = 0.0
            isav = -1
            redsav = 0.0
            temp = 0.0  # force scope outside i loop below since needed later
            iu = int(17 * angbd + 3.1)
            for i in range(iu):  # i = 0, ..., iu-1
                angt = angbd * float(i + 1) / float(iu)
                sth = 2.0 * angt / (1.0 + angt ** 2)
                temp = shs + angt * (angt * dhd - 2.0 * dhs)
                rednew = sth * (angt * dredg - sredg - 0.5 * sth * temp)
                if rednew > redmax:
                    redmax = rednew
                    isav = i
                    rdprev = redsav
                elif i == isav + 1:
                    rdnext = rednew
                redsav = rednew

            # Return if the reduction is zero. Otherwise, set the sine and cosine
            # of the angle of the alternative iteration, and calculate SDEC.
            if isav == -1:
                restart_alt_loop = False
                break  # quit inner label 120 loop and return results

            if isav < iu - 1:
                temp = (rdnext - rdprev) / (2.0 * redmax - rdprev - rdnext)
                angt = angbd * (float(isav + 1) + 0.5 * temp) / float(iu)

            cth = (1.0 - angt ** 2) / (1.0 + angt ** 2)
            sth = 2.0 * angt / (1.0 + angt ** 2)
            temp = shs + angt * (angt * dhd - 2.0 * dhs)
            sdec = sth * (angt * dredg - sredg - 0.5 * sth * temp)

            if sdec <= 0.0:
                restart_alt_loop = False
                break  # quit inner label 120 loop and return results

            # Update GNEW, D and HRED. If the angle of the alternative iteration
            # is restricted by a bound on a free variable, that variable is fixed
            # at the bound.
            gnew += (cth - 1.0) * hred + sth * hs
            d[xbdi == 0] = cth * d[xbdi == 0] + sth * s[xbdi == 0]
            dredg = np.dot(d[xbdi == 0], gnew[xbdi == 0])
            gredsq = sumsq(gnew[xbdi == 0])
            hred = cth * hred + sth * hs

            qred += sdec
            if iact is not None and isav == iu - 1:
                nact += 1
                xbdi[iact] = xsav
                restart_alt_loop = True
                break  # quit inner label 120 loop and restart alt iteration loop (label 100)

            if (sdec <= 0.01 * qred):
                restart_alt_loop = False
                break  # quit inner label 120 loop and return results
            continue  # back to inner label 120 loop

        # End inner label 120 loop

        if restart_alt_loop:
            continue
        else:
            break  # end outer loop and quit

    # End while True (label 100)
    return d_within_bounds(d, xopt, sl, su, xbdi), gnew


def d_within_bounds(d, xopt, sl, su, xbdi):
    # Used in TRSBOX, force d to be within bounds
    # In Fortran code, is at label 190
    xnew = np.maximum(np.minimum(xopt + d, su), sl)
    xnew[xbdi == -1] = sl[xbdi == -1]
    xnew[xbdi == 1] = su[xbdi == 1]
    d = xnew - xopt
    return d


def ball_step(x0, g, Delta):
    # Given initial point x0, take largest step in direction g allowed by ||x|| <= Delta
    # That is, solve
    #   ||x0 + alpha*g||^2 = Delta^2, alpha >= 0
    # Using this method, solution exists whenever ||x0|| <= Delta^2 [take alpha=0 if g=0]
    gdotx0 = np.dot(g, x0)
    gsqnorm = np.dot(g, g)
    x0sqnorm = np.dot(x0, x0)
    if sqrt(gsqnorm) < ZERO_THRESH:  # Error catching: if g=0, make no step
        return 0.0
    else:
        return (sqrt(gdotx0**2 + gsqnorm*(Delta**2 - x0sqnorm)) - gdotx0) / gsqnorm


def trsbox_linear(g, a, b, Delta, use_fortran=USE_FORTRAN):
    if use_fortran:
        return trustregion.solve(g, None, Delta,
                                 sl=np.minimum(a, -ZERO_THRESH),
                                 su=np.maximum(b, ZERO_THRESH),
                                 verbose_output=False)

    # Solve the convex program:
    #   min_x   g' * x
    #   s.t.   a <= x <= b
    #           ||x||^2 <= Delta^2
    # using an active-set type approach

    n = g.size
    x = np.zeros((n,))
    dirn = -g
    cons_dirns = []

    # If g[i] = 0, never step along this direction
    constant_directions = np.where(np.abs(dirn) < ZERO_THRESH)[0]
    dirn[constant_directions] = 0.0
    cons_dirns += list(constant_directions)

    for i in range(n):
        if np.linalg.norm(dirn) < ZERO_THRESH:
            return x
        alpha_unc = ball_step(x, dirn, Delta)
        xnew = x + alpha_unc * dirn
        # Check if hit box bounds
        on_box_bdry = False
        hit_upper = None
        idx_hit = None
        for j in range(n):
            if j in cons_dirns:
                continue  # only looking at unconstrained directions
            if xnew[j] <= a[j]:
                on_box_bdry = True
                hit_upper = False
                idx_hit = j
                break
            elif xnew[j] >= b[j]:
                on_box_bdry = True
                hit_upper = True
                idx_hit = j
                break

        if not on_box_bdry:
            return xnew  # unconstrained solution
        else:
            # Go as far as possible until hit box, then remove that direction from 'dirn'
            cons_dirns.append(idx_hit)  # new constrained direction
            alpha_con = ((b[idx_hit] if hit_upper else a[idx_hit]) - x[idx_hit]) / dirn[idx_hit]
            x = x + alpha_con * dirn
            x[idx_hit] = b[idx_hit] if hit_upper else a[idx_hit]  # force boundary exactly
            dirn[idx_hit] = 0.0  # no more searching this direction
    return x


def trsbox_geometry(xbase, c, g, lower, upper, Delta):
    # Given a Lagrange polynomial defined by: L(x) = c + g' * (x - xbase)
    # Maximise |L(x)| in a box + trust region - that is, solve:
    #   max_x  abs(c + g' * (x - xbase))
    #    s.t.  lower <= x <= upper
    #          ||x-xbase|| <= Delta
    # Setting s = x-xbase (or x = xbase + s), this is equivalent to:
    #   max_s  abs(c + g' * s)
    #   s.t.   lower - xbase <= s <= upper - xbase
    #          ||s|| <= Delta
    smin = trsbox_linear(g, lower - xbase, upper - xbase, Delta)  # minimise g' * s
    smax = trsbox_linear(-g, lower - xbase, upper - xbase, Delta)  # maximise g' * s
    if abs(c + np.dot(g, smin)) >= abs(c + np.dot(g, smax)):  # choose the one with largest absolute value
        return xbase + smin
    else:
        return xbase + smax



"""
Various useful functions


This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.

The development of this software was sponsored by NAG Ltd. (http://www.nag.co.uk)
and the EPSRC Centre For Doctoral Training in Industrially Focused Mathematical
Modelling (EP/L015803/1) at the University of Oxford. Please contact NAG for
alternative licensing.

"""

# Ensure compatibility with Python 2
from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import numpy as np
import sys


__all__ = ['sumsq', 'eval_objective', 'model_value',
           'random_directions_within_bounds', 'apply_scaling', 'remove_scaling',
           'random_orthog_directions_new_subspace_within_bounds']

def sumsq(x):
    # There are several ways to calculate sum of squares of a vector:
    #   np.dot(x,x)
    #   np.sum(x**2)
    #   np.sum(np.square(x))
    #   etc.
    # Using the timeit routine, it seems like dot(x,x) is ~3-4x faster than other methods
    return np.dot(x, x)


def eval_objective(objfun, x, args=(), verbose=True, eval_num=0, pt_num=None, full_x_thresh=6, check_for_overflow=True):
    # Evaluate objective function
    if check_for_overflow:
        try:
            f = objfun(x, *args)
        except OverflowError:
            f = sys.float_info.max
    else:
        f = objfun(x, *args)

    if verbose:
        if len(x) < full_x_thresh:
            if pt_num is None:
                logging.info("Function eval %i has f = %.15g at x = " % (eval_num, f) + str(x))
            else:
                logging.info("Function eval %i at point %i has f = %.15g at x = " % (eval_num, pt_num, f) + str(x))
        else:
            if pt_num is None:
                logging.info("Function eval %i has f = %.15g at x = [...]" % (eval_num, f))
            else:
                logging.info("Function eval %i at point %i has f = %.15g at x = [...]" % (eval_num, pt_num, f))

    return f


def model_value(g, hess, s):
    # Calculate model value (s^T * g + 0.5* s^T * H * s) = s^T * (gopt + 0.5 * H*s)
    assert g.shape == s.shape, "g and s have incompatible sizes"
    Hs = hess.dot(s)
    return np.dot(s, g + 0.5*Hs)


def get_scale(dirn, delta, lower, upper):
    scale = delta
    for j in range(len(dirn)):
        if dirn[j] < 0.0:
            scale = min(scale, lower[j] / dirn[j])
        elif dirn[j] > 0.0:
            scale = min(scale, upper[j] / dirn[j])
    return scale


def random_orthog_directions_new_subspace_within_bounds(num_pts, delta, lower, upper, Q=None, box_bound_thresh=0.01):
    # Generate num_pts random directions d1, d2, ...
    # so that lower <= d1 <= upper and ||d1|| ~ delta [perhaps not equal if constraint active]
    # Q is n*p matrix with orthonormal columns
    # Want to generate the points such that d is orthogonal to each column of Q (if possible)
    # Need num_pts <= n-p to maintain orthogonality
    # If within box_bound_thresh*delta of any bound, drop orthogonality requirement (too hard to do properly)
    n = len(lower)
    if Q is not None:
        p = Q.shape[1]
        assert Q.shape == (n, p), "Q must have n rows"
    else:
        p = 0
    assert lower.shape == (n,), "lower must be a vector"
    assert upper.shape == (n,), "lower and upper have incompatible sizes"
    assert np.min(upper) >= -1e-15, "upper must be non-negative"
    assert np.max(lower) <= 1e-15, "lower must be non-positive"
    assert np.min(upper - lower) > 0.0, "upper must be > lower"
    assert delta > 0, "delta must be strictly positive"
    assert num_pts > 0, "num_pts must be strictly positive"
    assert num_pts <= n - p, "num_pts must be <= n-p (p=number of columns of Q)"

    results = np.zeros((num_pts, n))  # save space for results

    # Check if near the boundary
    if np.min(upper) < box_bound_thresh*delta or np.max(lower) > box_bound_thresh*delta:
        return random_directions_within_bounds(num_pts, delta, lower, upper)

    # Otherwise, we are in the strict interior
    A = np.random.normal(size=(n, num_pts))
    if Q is not None:
        A = A - np.dot(Q, np.dot(Q.T, A))  # make orthogonal to columns of Q
    A_Q, _ = np.linalg.qr(A, mode='reduced')  # make directions orthonormal
    for i in range(num_pts):
        scale = get_scale(A_Q[:,i], delta, lower, upper)  # from above boundary check, scale should be >= 0.01
        results[i, :] = np.maximum(np.minimum(scale * A_Q[:, i], upper), lower)  # double-check bounds satisfied

    return results


def random_directions_within_bounds(num_pts, delta, lower, upper):
    # Generate num_pts random directions d1, d2, ...
    # so that lower <= d1 <= upper and ||d1|| ~ delta [perhaps not equal if constraint active]
    # Directions should be completely random (as much as possible while staying within bounds)
    n = len(lower)
    assert lower.shape == (n,), "lower must be a vector"
    assert upper.shape == (n,), "lower and upper have incompatible sizes"
    assert np.min(upper) >= -1e-15, "upper must be non-negative"
    assert np.max(lower) <= 1e-15, "lower must be non-positive"
    assert np.min(upper - lower) > 0.0, "upper must be > lower"
    assert delta > 0, "delta must be strictly positive"
    assert num_pts > 0, "num_pts must be strictly positive"
    results = np.zeros((n, num_pts))  # save space for results
    # Find the active set
    idx_l = (lower == 0)
    idx_u = (upper == 0)
    active = np.logical_or(idx_l, idx_u)
    # inactive = np.logical_not(active)
    nactive = np.sum(active)
    # ninactive = n - nactive
    idx_active = np.where(active)[0]  # indices of active constraints
    for i in range(num_pts):
        dirn = np.random.normal(size=(n,))
        for j in range(nactive):
            idx = idx_active[j]
            sign = 1.0 if idx_l[idx] else -1.0  # desired sign of direction shift
            if dirn[idx]*sign < 0.0:
                dirn[idx] *= -1.0
        dirn = dirn / np.linalg.norm(dirn)
        scale = get_scale(dirn, delta, lower, upper)
        results[:, i] = dirn * scale
    # Finally, scale by delta and make sure everything is within bounds
    for i in range(num_pts):
        results[:, i] = np.maximum(np.minimum(results[:, i], upper), lower)
    return results.T


def apply_scaling(x_raw, scaling_changes):
    if scaling_changes is None:
        return x_raw
    shift, scale = scaling_changes
    return (x_raw - shift) / scale


def remove_scaling(x_scaled, scaling_changes):
    if scaling_changes is None:
        return x_scaled
    shift, scale = scaling_changes
    return shift + x_scaled * scale




"""
Block-Hess
=====

Block-hess is a high-dimensional derivative-free solver.

It solves the unconstrained nonlinear minimisation problem:
    min_{x}  f(x)

----

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.

The development of this software was sponsored by NAG Ltd. (http://www.nag.co.uk)
and the EPSRC Centre For Doctoral Training in Industrially Focused Mathematical
Modelling (EP/L015803/1) at the University of Oxford. Please contact NAG for
alternative licensing.


"""

# Ensure compatibility with Python 2
from __future__ import absolute_import, division, print_function, unicode_literals

__version__ = '1.0'
__all__ = ['__version__']

from .solver import solve
__all__ += ['solve']


"""
Diagnostic Info
====

A class containing diagnostic information (optionally) produced by the solver.


This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.

The development of this software was sponsored by NAG Ltd. (http://www.nag.co.uk)
and the EPSRC Centre For Doctoral Training in Industrially Focused Mathematical
Modelling (EP/L015803/1) at the University of Oxford. Please contact NAG for
alternative licensing.

"""

# Ensure compatibility with Python 2
from __future__ import absolute_import, division, print_function, unicode_literals

import numpy as np
import pandas as pd


__all__ = ['DiagnosticInfo', 'ITER_VERY_SUCCESSFUL', 'ITER_SUCCESSFUL', 'ITER_ACCEPTABLE',
           'ITER_UNSUCCESSFUL', 'ITER_SAFETY']


ITER_VERY_SUCCESSFUL = "Very successful"   # ratio >= 0.7, no geometry update
ITER_SUCCESSFUL = "Successful"  # 0.1 <= ratio < 0.7, no geometry update
ITER_ACCEPTABLE = "Acceptable"  # 0 <= ratio < 0.1
# ITER_ACCEPTABLE_GEOM = "Acceptable (geom fixed)"  # 0 <= ratio < 0.1, with geometry update
# ITER_ACCEPTABLE_NO_GEOM = "Acceptable (geom not fixed)"  # 0 <= ratio < 0.1, without geometry update
ITER_UNSUCCESSFUL = "Unsuccessful"  # ratio < 0
# ITER_UNSUCCESSFUL_GEOM = "Unsuccessful (geom fixed)"  # ratio < 0, with geometry update
# ITER_UNSUCCESSFUL_NO_GEOM = "Unsuccessful (geom not fixed)"  # ratio < 0, without geometry update (possibly rho reduced)
ITER_SAFETY = "Safety"  # safety step taken (||s|| too small compared to rho)
ITER_INIT = "Initial setup"


def srank(A):  # stable rank of a matrix
    if np.linalg.norm(A, ord=2) < 1e-16:
        return 0.0
    else:
        return np.linalg.norm(A, ord='fro')**2 / np.linalg.norm(A, ord=2)**2


def cond(A, p=2):  # condition number for any matrix
    # if A.shape[0] == A.shape[1]:
    #     return np.linalg.cond(A, p=p)
    # else:
    try:
        return np.linalg.norm(A, ord=p) * np.linalg.norm(np.linalg.pinv(A), ord=p)
    except np.linalg.LinAlgError:
        return 1e15  # can't compute pseudoinverse, assume ill conditioned


class DiagnosticInfo:
    def __init__(self, x0, f0, delta0, rho0, nf, with_xk=False, with_poisedness=False, with_rho=False):  # initialise with first info
        self.data = {}
        # Initialise everything we want to store
        self.save_xk = with_xk
        self.save_poisedness = with_poisedness
        if self.save_xk:
            self.data["xk"] = [x0]
        self.data["fk"] = [f0]

        self.data["delta"] = [delta0]
        self.save_rho = with_rho
        if self.save_rho:
            self.data["rho"] = [rho0]

        self.data["H_norm"] = [np.nan]
        self.data["H_norm2"] = [np.nan]
        self.data["H_cond"] = [np.nan]

        self.data["norm_gk"] = [np.nan]
        self.data["norm_sk"] = [np.nan]

        self.data["nruns"] = [1]
        self.data["nf"] = [nf]
        # self.data["nx"] = [nx]
        self.data["iter_this_run"] = [0]
        self.data["iters_total"] = [0]

        self.data["iter_type"] = [ITER_INIT]
        self.data["ratio"] = [np.nan]

        self.data["ndirs"] = [np.nan]  # number of directions used to construct model
        self.data["npt"] = [np.nan]  # number of points used to construct model
        if self.save_poisedness:
            self.data["poisedness"] = [np.nan]  # poisedness of model in B(xk,delta) (reduced space) used to compute trust region step

        # And some things will be constant throughout
        self.data["n"] = [len(x0)]
        return

    def to_dataframe(self):
        data_to_save = {}
        for key in self.data:
            if (key == "xk" and not self.save_xk):
                continue  # skip
            data_to_save[key] = self.data[key]
        return pd.DataFrame(data_to_save)

    def to_csv(self, filename):
        df = self.to_dataframe()
        df.to_csv(filename)

    def save_info(self, model, delta, rho, H, ndirs, npt, norm_sk, norm_gk, nruns, nf, iter_this_run, iter_type, ratio, model_poisedness):
        if self.save_xk:
            self.data["xk"].append(model.xopt())
        self.data["fk"].append(model.fopt())

        self.data["delta"].append(delta)
        if self.save_rho:
            self.data["rho"].append(rho)

        if H is not None:
            self.data["H_norm"].append(np.linalg.norm(H, ord='fro'))
            self.data["H_norm2"].append(np.linalg.norm(H, ord=2))
            self.data["H_cond"].append(cond(H, p=2))
        else:  # linear models
            self.data["H_norm"].append(0.0)
            self.data["H_norm2"].append(0.0)
            self.data["H_cond"].append(0.0)

        self.data["norm_gk"].append(norm_gk)
        self.data["norm_sk"].append(norm_sk)

        self.data["nruns"].append(nruns)
        self.data["nf"].append(nf)
        # self.data["nx"].append(nx)
        self.data["iter_this_run"].append(iter_this_run)
        self.data["iters_total"].append(len(self.data["iters_total"]))

        self.data["iter_type"].append(iter_type)
        self.data["ratio"].append(ratio)

        self.data["ndirs"].append(ndirs)
        self.data["npt"].append(npt)
        if self.save_poisedness:
            self.data["poisedness"].append(model_poisedness)

        # Constant stuff - carry over previous value
        for key in ["n"]:
            self.data[key].append(self.data[key][-1])
        return



"""
Wrapper for exit information

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.

The development of this software was sponsored by NAG Ltd. (http://www.nag.co.uk)
and the EPSRC Centre For Doctoral Training in Industrially Focused Mathematical
Modelling (EP/L015803/1) at the University of Oxford. Please contact NAG for
alternative licensing.
"""

# Ensure compatibility with Python 2
from __future__ import absolute_import, division, print_function, unicode_literals

import numpy as np


__all__ = ['EXIT_AUTO_DETECT_RESTART_WARNING', 'EXIT_FALSE_SUCCESS_WARNING', 'EXIT_SLOW_WARNING', 'EXIT_MAXFUN_WARNING',
           'EXIT_SUCCESS', 'EXIT_INPUT_ERROR', 'EXIT_TR_INCREASE_ERROR', 'EXIT_LINALG_ERROR', 'ExitInformation',
           'OptimResults']


EXIT_AUTO_DETECT_RESTART_WARNING = 4  # warning, auto-detected restart criteria
EXIT_FALSE_SUCCESS_WARNING = 3  # warning, maximum fake successful steps reached
EXIT_SLOW_WARNING = 2  # warning, maximum number of slow (successful) iterations reached
EXIT_MAXFUN_WARNING = 1  # warning, reached max function evals
EXIT_SUCCESS = 0  # successful finish (rho=rhoend, sufficient objective reduction, or everything in noise level)
EXIT_INPUT_ERROR = -1  # error, bad inputs
EXIT_TR_INCREASE_ERROR = -2  # error, trust region step increased model value
EXIT_LINALG_ERROR = -3  # error, linalg error (singular matrix encountered)
# EXIT_ALTMOV_MEMORY_ERROR = -4  # error, stpsav issue in ALTMOV


class ExitInformation:
    def __init__(self, flag, msg_details):
        self.flag = flag
        self.msg = msg_details

    def flag(self):
        return self.flag

    def message(self, with_stem=True):
        if not with_stem:
            return self.msg
        elif self.flag == EXIT_SUCCESS:
            return "Success: " + self.msg
        elif self.flag == EXIT_SLOW_WARNING:
            return "Warning (slow progress): " + self.msg
        elif self.flag == EXIT_MAXFUN_WARNING:
            return "Warning (max evals): " + self.msg
        elif self.flag == EXIT_INPUT_ERROR:
            return "Error (bad input): " + self.msg
        elif self.flag == EXIT_TR_INCREASE_ERROR:
            return "Error (trust region increase): " + self.msg
        elif self.flag == EXIT_LINALG_ERROR:
            return "Error (linear algebra): " + self.msg
        elif self.flag == EXIT_FALSE_SUCCESS_WARNING:
            return "Warning (max false good steps): " + self.msg
        # elif self.flag == EXIT_ALTMOV_MEMORY_ERROR:
        #     return "Error (geometry step): " + self.msg
        else:
            return "Unknown exit flag: " + self.msg

    def able_to_do_restart(self):
        if self.flag in [EXIT_TR_INCREASE_ERROR, EXIT_LINALG_ERROR, EXIT_SLOW_WARNING, EXIT_AUTO_DETECT_RESTART_WARNING]:
            return True
        elif self.flag in [EXIT_MAXFUN_WARNING, EXIT_INPUT_ERROR]:
            return False
        else:
            # Successful step (rho=rhoend, noise level termination, or value small)
            return "sufficiently small" not in self.msg  # restart for rho=rhoend and noise level termination


# A container for the results of the optimization routine
class OptimResults(object):
    def __init__(self, xmin, fmin, gradmin, hessmin, nf, nruns, exit_flag, exit_msg):
        self.x = xmin
        self.f = fmin
        self.gradient = gradmin
        self.hessian = hessmin
        self.nf = nf
        # self.nx = nx
        self.nruns = nruns
        self.flag = exit_flag
        self.msg = exit_msg
        self.diagnostic_info = None
        # Set standard names for exit flags
        self.EXIT_SLOW_WARNING = EXIT_SLOW_WARNING
        self.EXIT_MAXFUN_WARNING = EXIT_MAXFUN_WARNING
        self.EXIT_SUCCESS = EXIT_SUCCESS
        self.EXIT_INPUT_ERROR = EXIT_INPUT_ERROR
        self.EXIT_TR_INCREASE_ERROR = EXIT_TR_INCREASE_ERROR
        self.EXIT_LINALG_ERROR = EXIT_LINALG_ERROR
        self.EXIT_FALSE_SUCCESS_WARNING = EXIT_FALSE_SUCCESS_WARNING

    def __str__(self):
        # Result of calling print(soln)
        output = "****** RSDFO-Q Results ******\n"
        if self.flag != self.EXIT_INPUT_ERROR:
            if np.size(self.x) < 200:
                output += "Solution xmin = %s\n" % str(self.x)
            else:
                output += "Not showing solution because it is too long; check self.xmin\n"
            output += "Objective value f(xmin) = %.10g\n" % self.f
            # output += "Needed %g objective evaluations (at %g points)\n" % (self.nf, self.nx)
            output += "Needed %g objective evaluations\n" % (self.nf)
            if self.nruns > 1:
                output += "Did a total of %g runs\n" % self.nruns
            if self.gradient is not None and np.size(self.gradient) < 100:
                output += "Approximate gradient = %s\n" % str(self.gradient)
            elif self.gradient is not None:
                output += "Not showing approximate gradient because it is too long; check self.gradient\n"
            if self.hessian is not None and np.size(self.hessian) < 200:
                output += "Approximate Hessian = %s\n" % str(self.hessian)
            elif self.hessian is not None:
                output += "Not showing approximate Hessian because it is too long; check self.hessian\n"
            if self.diagnostic_info is not None:
                output += "Diagnostic information available; check self.diagnostic_info\n"
        output += "Exit flag = %g\n" % self.flag
        output += "%s\n" % self.msg
        output += "****************************\n"
        return output


"""
Main solver
===========

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.

The development of this software was sponsored by NAG Ltd. (http://www.nag.co.uk)
and the EPSRC Centre For Doctoral Training in Industrially Focused Mathematical
Modelling (EP/L015803/1) at the University of Oxford. Please contact NAG for
alternative licensing.
"""

# Ensure compatibility with Python 2
from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import numpy as np
import warnings

from .diagnostic_info import *
from .exit_information import *
from .model import *
from .params import ParameterList
from .trust_region import trsbox, trsbox_linear
from .util import sumsq, apply_scaling, remove_scaling, eval_objective, \
    random_orthog_directions_new_subspace_within_bounds


__all__ = ['solver']


def min_npt(p):
    return p + 1


def max_npt(p):
    return (p + 1) * (p + 2) // 2


def trust_region_step(gk, Hk, delta):
    # Light trust-region subproblem wrapper (for problems without bound constraints)
    p = len(gk)
    xl = np.full((p,), -1e20)
    xu = np.full((p,), 1e20)
    if Hk is not None:
        sk_red, _, crvmin = trsbox(np.zeros((p,)), gk, Hk, xl, xu, delta)  # step in reduced space
    else:
        sk_red = trsbox_linear(gk, xl, xu, delta)
        crvmin = 0.0
    return sk_red, crvmin


def update_tr(delta, rho, ratio, norm_sk, params):
    if params("tr_radius.update_method") == 'bobyqa':
        if ratio < params("tr_radius.eta1"):  # ratio < 0.1
            iter_type = ITER_ACCEPTABLE if ratio > 0.0 else ITER_UNSUCCESSFUL
            delta = min(params("tr_radius.gamma_dec") * delta, norm_sk)

        elif ratio <= params("tr_radius.eta2"):  # 0.1 <= ratio <= 0.7
            iter_type = ITER_SUCCESSFUL
            delta = max(params("tr_radius.gamma_dec") * delta, norm_sk)

        else:  # (ratio > eta2 = 0.7)
            iter_type = ITER_VERY_SUCCESSFUL
            delta = min(max(params("tr_radius.gamma_inc") * delta, params("tr_radius.gamma_inc_overline") * norm_sk),
                        params("tr_radius.delta_max"))

    else:
        raise RuntimeError("Unknown TR updating method: %s" % params("tr_radius.update_method"))

    if params("tr_radius.use_rho"):
        delta = max(delta, rho)
    return delta, iter_type


def done_with_current_rho(rho, last_successful_iter, current_iter, diffs, crvmin, in_safety=False):
    # crvmin comes from trust region step

    # Wait at least 3 iterations between reductions of rho
    if current_iter <= last_successful_iter + 5:
        return False

    errbig = max(diffs)
    frhosq = 0.125 * rho ** 2
    # if in_safety and crvmin > 0.0 and errbig > frhosq * crvmin:
    #     logging.debug("Not reducing because of this (crvmin = %g)" % crvmin)
    #     return False

    # Otherwise...
    return True


def reduce_rho(delta, rho, rhoend, params):
    if delta > 1.5*rho:  # nothing needed if delta > rho
        return delta, rho

    alpha1 = params("tr_radius.alpha1")
    alpha2 = params("tr_radius.alpha2")
    ratio = rho / rhoend
    if ratio <= 16.0:
        new_rho = rhoend
    elif ratio <= 250.0:
        new_rho = np.sqrt(ratio) * rhoend  # geometric average of rho and rhoend
    else:
        new_rho = alpha1 * rho

    new_delta = max(alpha2 * rho, new_rho)  # self.rho = old rho
    return new_delta, new_rho


def eval_objfun(model, x, objfun, args, scaling_changes, nf, maxfun, params):
    # Simple wrapper to objective evaluation, returning exit codes as needed
    nf += 1
    f = eval_objective(objfun, remove_scaling(x, scaling_changes), args=args,
                          eval_num=nf, full_x_thresh=params("logging.n_to_print_whole_x_vector"),
                          check_for_overflow=params("general.check_objfun_for_overflow"))

    if f <= model.min_objective_value():
        model.save_point(x, f)  # save, since this point was an improvement
        exit_info = ExitInformation(EXIT_SUCCESS, "Objective is sufficiently small")

    elif nf >= maxfun:
        model.save_point(x, f)  # save, just in case this point was an improvement
        exit_info = ExitInformation(EXIT_MAXFUN_WARNING, "Objective has been called MAXFUN times")

    else:
        exit_info = None

    return f, exit_info, nf


def solve_main(objfun, x0, args, rhobeg, rhoend, maxfun, nruns_so_far, nf_so_far,
               params, scaling_changes, fixed_block_size, hess_npt):
    # First, evaluate objfun at x0 - this gives us m
    logging.debug("Block size %g with %g Hessian points" % (fixed_block_size, hess_npt))
    nf = nf_so_far + 1
    f0 = eval_objective(objfun, remove_scaling(x0, scaling_changes), args=args, eval_num=nf,
                        full_x_thresh=params("logging.n_to_print_whole_x_vector"),
                        check_for_overflow=params("general.check_objfun_for_overflow"))
    exit_info = None

    if f0 <= params("model.abs_tol"):
        exit_info = ExitInformation(EXIT_SUCCESS, "Objective is sufficiently small")

    if exit_info is not None:
        return x0, f0, None, None, nf, nruns_so_far + 1, exit_info, None

    # Initialize model
    delta = rhobeg
    rho = rhobeg if params("tr_radius.use_rho") else 0.0
    model = InterpSet(fixed_block_size, hess_npt, x0, f0, n=len(x0),
                      abs_tol=params("model.abs_tol"), precondition=params("model.precondition_linear_system"))
    exit_info, nf = model.initialise_interp_simplex(delta, objfun, args, scaling_changes, nf, maxfun,
                                                    use_coord_directions=params("geometry.use_coord_directions"),
                                                    box_bound_thresh=params("geometry.direcion_box_bound_thresh"),
                                                    full_x_thresh=params("logging.n_to_print_whole_x_vector"),
                                                    check_for_overflow=params("general.check_objfun_for_overflow"))
    if exit_info is not None:
        xopt, fopt = model.get_final_results()
        return xopt, fopt, None, None, nf, nruns_so_far + 1, exit_info, None

    if params("logging.save_diagnostic_info"):
        diagnostic_info = DiagnosticInfo(x0, f0, delta, rho, nf, with_xk=params("logging.save_xk"),
                                         with_poisedness=params("logging.save_poisedness"),
                                         with_rho=params("tr_radius.use_rho"))
    else:
        diagnostic_info = None

    current_iter = 0
    last_fopts_for_slowterm = []
    num_consecutive_slow_iters = 0
    # Adaptive checking - how many safety/very successful steps have we had in a row so far?
    num_consecutive_safety_steps = 0
    H_to_save, Q_to_save = None, None  # previous model Hessian (for minimal change)
    last_successful_iter = 0  # determines when we can reduce rho
    diffs = [0.0, 0.0, 0.0]

    while True:
        current_iter += 1
        if params("tr_radius.use_rho"):
            logging.debug("*** Iter %g (delta = %g, rho = %g) ***" % (current_iter, delta, rho))
        else:
            logging.debug("*** Iter %g (delta = %g) ***" % (current_iter, delta))
        # print("***** Iter %g *****" % current_iter)

        # Don't bother sampling new points after a safety step - we just reduce delta and try again
        recycle_model = (0 < num_consecutive_safety_steps <= params("geometry.safety_steps_before_redraw"))
        if recycle_model:
            logging.debug("Recycling model from previous iteration (safety step)")

        model.factorise_simplex_system()
        logging.debug("Using model with %g points in p=%g directions" % (model.p+1+model.num_hess_pts, model.p))

        xk = model.xopt()
        fk = model.fopt()
        if params("adaptive.min_change_hess") and H_to_save is not None:
            # Switch previous Hessian to new basis
            # H_to_save is p*p in basis given by Q_to_save, h_old is p*p in basis given by model.Q
            basis_switch = np.dot(Q_to_save.T, model.Q)  # cost O(np^2) to compute, size p*p
            H_old = np.dot(basis_switch.T, np.dot(H_to_save, basis_switch))  # all matrices p*p, cost O(p^3) to compute
            H_old = 0.5*(H_old + H_old.T)  # force symmetry, avoiding rounding errors
        else:
            H_old = None
        interp_ok, ck, gk, Hk = model.build_quadratic_interp_model(Hprev=H_old, model_in_full_space=False)  # based at xopt, in reduced space
        ndirs_used = model.p
        npt_used = model.p + 1 + model.num_hess_pts
        delta_used = delta
        rho_used = rho  # for diagnostic info only
        if not interp_ok:
            exit_info = ExitInformation(EXIT_LINALG_ERROR, "Failed to build interpolation model")
            break  # quit

        # Save Hessian for next iteration (minimal change)
        if params("adaptive.min_change_hess"):
            Q_to_save = model.Q.copy()
            H_to_save = Hk.copy() if Hk is not None else None

        # (optionally) save poisedness of interpolation model
        poisedness = model.poisedness_in_reduced_space(delta) if params("logging.save_poisedness") else None

        # Calculate step/predicted reduction
        sk_red, crvmin = trust_region_step(gk, Hk, delta)  # step in reduced space
        sk_full = model.project_to_full_space(sk_red)
        norm_sk = np.linalg.norm(sk_red)
        xnew = xk + sk_full
        if Hk is not None:
            pred_reduction = -np.dot(sk_red, gk + 0.5 * Hk.dot(sk_red))
        else:
            pred_reduction = -np.dot(sk_red, gk)

        if norm_sk < params("general.safety_step_thresh") * (rho if params("tr_radius.use_rho") else delta) \
                or pred_reduction < 0.0:  # step too short or TRS gave model increase
            logging.debug("Safety step")
            iter_type = ITER_SAFETY
            num_consecutive_safety_steps += 1
            # num_consecutive_very_successful_steps = 0
            ratio = None  # used for diagnostic info only

            if params("tr_radius.use_rho"):
                delta = max(params("tr_radius.gamma_dec") * delta, rho)
                if not done_with_current_rho(rho, last_successful_iter, current_iter, diffs, crvmin,
                                             in_safety=True) or delta > rho:
                    # Delete a bad point (equivalent to fixing geometry)
                    try:
                        sigmas = model.poisedness_of_each_simplex_point(d=sk_red, d_in_full_space=False)
                        sqdists = model.simplex_distances_to_xopt(include_kopt=True)  # ||yt-xk||^2
                        vals = sigmas * np.maximum(sqdists ** 2 / delta ** 4, 1)  # BOBYQA point to remove criterion
                        vals[model.kopt] = -1.0  # make sure kopt is never selected
                        k = np.argmax(vals)
                    except np.linalg.LinAlgError:
                        # If poisedness calculation fails, revert to dropping furthest points
                        sqdists = model.simplex_distances_to_xopt(include_kopt=True)  # ||yt-xk||^2
                        k = np.argmax(sqdists)

                    model.remove_simplex_point(k, check_not_kopt=True)

                    # if params("general.use_safety_step") and norm_sk > rho:
                    #     last_successful_iter = current_iter
                else:
                    delta, rho = reduce_rho(delta, rho, rhoend, params)
                    last_successful_iter = current_iter

                if rho <= rhoend:
                    exit_info = ExitInformation(EXIT_SUCCESS, "rho has reached rhoend")
                    break  # quit
            else:
                delta = params("tr_radius.gamma_dec") * delta
                if delta <= rhoend:
                    exit_info = ExitInformation(EXIT_SUCCESS, "delta has reached rhoend")
                    break  # quit
            # (end of safety step)
        else:
            # (start of normal step)
            num_consecutive_safety_steps = 0

            # Evaluate objective at xnew
            fnew, exit_info, nf = eval_objfun(model, xnew, objfun, args, scaling_changes, nf, maxfun, params)
            if exit_info is not None:
                break  # quit

            # Slow termination checking
            if fnew < fk:
                if len(last_fopts_for_slowterm) <= params("slow.history_for_slow"):
                    last_fopts_for_slowterm.append(fk)
                    num_consecutive_slow_iters = 0
                else:
                    last_fopts_for_slowterm = last_fopts_for_slowterm[1:] + [fk]
                    this_iter_slow = (np.log(last_fopts_for_slowterm[0]) - np.log(fk)) / float(
                        params("slow.history_for_slow")) < params("slow.thresh_for_slow")
                    if this_iter_slow:
                        num_consecutive_slow_iters += 1
                    else:
                        num_consecutive_slow_iters = 0

                # Do slow termination check
                if num_consecutive_slow_iters >= params("slow.max_slow_iters"):
                    model.save_point(xnew, fnew)  # save, since this point was an improvement
                    exit_info = ExitInformation(EXIT_SLOW_WARNING, "Maximum slow iterations reached")
                    break  # quit

            # Decide on type of step
            actual_reduction = fk - fnew
            ratio = actual_reduction / pred_reduction
            if min(norm_sk, delta) > rho:  # if ||sk|| >= rho, successful!
                last_successful_iter = current_iter
            diffs = [abs(actual_reduction - pred_reduction), diffs[0], diffs[1]]

            # Update trust region radius
            delta, iter_type = update_tr(delta, rho, ratio, norm_sk, params)

            logging.debug("Ratio = %g (%s)" % (ratio, iter_type))

            # Add xnew to interpolation set
            if model.p < model.n:
                logging.debug("Appending xk+sk")
                model.append_simplex_point(xnew, fnew)  # updates xopt
                xnew_appended = True
            else:
                # If the model is full, replace xnew with the point furthest from xk (from previous iteration)
                try:
                    sigmas = model.poisedness_of_each_simplex_point(d=sk_red, d_in_full_space=False)
                    sqdists = model.simplex_distances_to_xopt(include_kopt=True)  # ||yt-xk||^2
                    vals = sigmas * np.maximum(sqdists ** 2 / delta ** 4, 1)  # BOBYQA point to remove criterion
                    vals[model.kopt] = -1.0  # make sure kopt is never selected
                    knew = np.argmax(vals)
                except np.linalg.LinAlgError:
                    # If poisedness calculation fails, revert to dropping furthest points
                    sqdists = model.simplex_distances_to_xopt(include_kopt=True)  # ||yt-xk||^2
                    knew = np.argmax(sqdists)
                logging.debug("Changing simplex point %g to xk+sk" % knew)
                model.change_simplex_point(knew, xnew, fnew, check_not_kopt=True)  # updates xopt
                xnew_appended = False

            # Drop points no longer needed
            # Remove at least 1 from xnew (if appended) and 1 to make space for a new direction in next iter
            min_npt_to_drop = (2 if xnew_appended else 1)
            p_drop = (1.0 - params("adaptive.simplex_save_frac_successful")) * model.p
            if iter_type in [ITER_ACCEPTABLE, ITER_UNSUCCESSFUL]:
                p_drop = (1.0 - params("adaptive.simplex_save_frac_unsuccessful")) * model.p
                if params("hessian.forget_on_unsuccessful"):
                    H_to_save = None
                if params("hessian.drop_interp_points_on_unsuccessful"):
                    model.clear_hessian_points()
            ndirs_to_keep = max(0, min(model.p - int(p_drop), model.p - min_npt_to_drop))  # TODO this is correct
            # ndirs_to_keep = max(0, min(int(p_drop), model.p - min_npt_to_drop))
            ndirs_to_drop = model.p - ndirs_to_keep
            logging.debug("After adding xnew, model has %g directions, dropping %g" % (model.p, ndirs_to_drop))

            # Criteria of points to remove:
            try:
                sigmas = model.poisedness_of_each_simplex_point(delta=delta)
                sqdists = model.simplex_distances_to_xopt(include_kopt=True)  # ||yt-xk||^2
                vals = sigmas * np.maximum(sqdists ** 2 / delta ** 4, 1)  # BOBYQA point to remove criterion
            except np.linalg.LinAlgError:
                # If poisedness calculation fails, revert to dropping furthest points
                logging.debug("Poisedness failed, using distance only")
                vals = model.simplex_distances_to_xopt(include_kopt=True)  # ||yt-xk||^2
            vals[model.kopt] = -np.inf  # make sure kopt is never selected
            # logging.debug("kopt = %g, argmin = %g" % (model.kopt, np.argmin(model.fvals)))
            # logging.debug("Here, fvals in [%.20g, %.20g]" % (np.min(model.fvals), np.max(model.fvals)))

            for i in range(ndirs_to_drop):
                k = np.argmax(vals)
                # logging.debug("Dropping %g (val = %g) // kopt = %g with val = %g" % (k, vals[k], model.kopt, vals[model.kopt]))
                vals = np.delete(vals, k)  # keep vals indices in line with indices of model.points
                model.remove_simplex_point(k, check_not_kopt=True)
                # try:
                #     model.remove_simplex_point(k, check_not_kopt=True)
                #     # logging.debug("Model kopt = %g, with val = %g" % (model.kopt, vals[model.kopt]))
                # except AssertionError as e:
                #     for i in range(model.p + 1):
                #         logging.debug("%g: %g" % (i, vals[i]))
                #     raise e

            if params("tr_radius.use_rho"):
                if ratio < 0 and done_with_current_rho(rho, last_successful_iter, current_iter, diffs, crvmin,
                                                       in_safety=False) and delta <= rho:
                    delta, rho = reduce_rho(delta, rho, rhoend, params)
                    last_successful_iter = current_iter

                if rho <= rhoend:
                    exit_info = ExitInformation(EXIT_SUCCESS, "rho has reached rhoend")
                    break  # quit
            else:
                if delta <= rhoend:
                    exit_info = ExitInformation(EXIT_SUCCESS, "delta has reached rhoend")
                    break  # quit
            # (end of normal step)

        # Add new simplex points to get back to correct size - do in both safety and normal steps
        ndirs_to_add = fixed_block_size - model.p
        logging.debug("Model has p=%g, adding %g points" % (model.p, ndirs_to_add))
        for i in range(ndirs_to_add):
            model.factorise_simplex_system()  # now can get model.Q (None if no directions stored currently, which is ok for next line)
            d = random_orthog_directions_new_subspace_within_bounds(1, delta, model.xl - model.xopt(),
                                                                    model.xu - model.xopt(),
                                                                    Q=model.Q,
                                                                    box_bound_thresh=params(
                                                                        "geometry.direcion_box_bound_thresh"))[0, :]

            # Evaluate objective
            xnew = model.xopt() + d
            fnew, exit_info, nf = eval_objfun(model, xnew, objfun, args, scaling_changes, nf, maxfun, params)
            if exit_info is not None:
                break  # quit (inner loop only)

            # Append xnew to model
            model.append_simplex_point(xnew, fnew)
        if exit_info is not None:
            break  # quit main loop
        logging.debug("At end of iteration, now have %g directions" % (model.p))

        # (in both safety and normal steps, update diagnostic info)
        if params("logging.save_diagnostic_info"):
            diagnostic_info.save_info(model, delta_used, rho_used, Hk, ndirs_used, npt_used, norm_sk, np.linalg.norm(gk),
                                      nruns_so_far + 1, nf, current_iter, iter_type, ratio, poisedness)
        continue

    xopt, fopt = model.get_final_results()
    return xopt, fopt, None, None, nf, nruns_so_far + 1, exit_info, diagnostic_info  # TODO no gradient/Hessian returned?


def solve(objfun, x0, args=(), fixed_block_size=None, hess_npt=None, bounds=None, rhobeg=None, rhoend=1e-8, maxfun=None,
          user_params=None, objfun_has_noise=False, scaling_within_bounds=False):
    n = len(x0)

    if fixed_block_size is not None:
        assert 1 <= fixed_block_size <= n, "fixed_block_size, if specified, must be in [1..n]"
    else:
        fixed_block_size = n  # default to full block

    if hess_npt is not None:
        assert min_npt(fixed_block_size) <= fixed_block_size + 1 + hess_npt <= max_npt(fixed_block_size), \
            "Invalid choice of hess_npt %g (%g points outside [%g,%g] for block size %g)" \
            % (hess_npt, fixed_block_size + 1 + hess_npt, min_npt(fixed_block_size), max_npt(fixed_block_size), fixed_block_size)
    else:
        hess_npt = fixed_block_size  # default to 2p+1 interpolation points in total

    # Set missing inputs (if not specified) to some sensible defaults
    if bounds is None:
        xl = None
        xu = None
        scaling_within_bounds = False
    else:
        raise RuntimeError(
            'Block Py-BOBYQA does not support bounds yet (needs update to trust region subproblem solver)')
        # assert len(bounds) == 2, "bounds must be a 2-tuple of (lower, upper), where both are arrays of size(x0)"
        # xl = bounds[0]
        # xu = bounds[1]

    if xl is None:
        xl = np.full((n,), -1e20)  # unconstrained
    if xu is None:
        xu = np.full((n,), 1e20)  # unconstrained

    if rhobeg is None:
        rhobeg = 0.1 if scaling_within_bounds else 0.1 * max(np.max(np.abs(x0)), 1.0)
    if maxfun is None:
        maxfun = min(100 * (n + 1), 1000)  # 100 gradients, capped at 1000

    # Set parameters
    params = ParameterList(int(n), int(maxfun), objfun_has_noise=objfun_has_noise)  # make sure int, not np.int
    if user_params is not None:
        for (key, val) in user_params.items():
            params(key, new_value=val)

    scaling_changes = None
    if scaling_within_bounds:
        shift = xl.copy()
        scale = xu - xl
        scaling_changes = (shift, scale)

    x0 = apply_scaling(x0, scaling_changes)
    xl = apply_scaling(xl, scaling_changes)
    xu = apply_scaling(xu, scaling_changes)

    exit_info = None
    # Input & parameter checks
    if exit_info is None and rhobeg < 0.0:
        exit_info = ExitInformation(EXIT_INPUT_ERROR, "rhobeg must be strictly positive")

    if exit_info is None and rhoend < 0.0:
        exit_info = ExitInformation(EXIT_INPUT_ERROR, "rhoend must be strictly positive")

    if exit_info is None and rhobeg <= rhoend:
        exit_info = ExitInformation(EXIT_INPUT_ERROR, "rhobeg must be > rhoend")

    if exit_info is None and maxfun <= 0:
        exit_info = ExitInformation(EXIT_INPUT_ERROR, "maxfun must be strictly positive")

    if exit_info is None and np.shape(x0) != (n,):
        exit_info = ExitInformation(EXIT_INPUT_ERROR, "x0 must be a vector")

    if exit_info is None and np.shape(x0) != np.shape(xl):
        exit_info = ExitInformation(EXIT_INPUT_ERROR, "lower bounds must have same shape as x0")

    if exit_info is None and np.shape(x0) != np.shape(xu):
        exit_info = ExitInformation(EXIT_INPUT_ERROR, "upper bounds must have same shape as x0")

    if exit_info is None and np.min(xu - xl) < 2.0 * rhobeg:
        exit_info = ExitInformation(EXIT_INPUT_ERROR, "gap between lower and upper must be at least 2*rhobeg")

    if maxfun <= n:
        warnings.warn("maxfun <= n: Are you sure your budget is large enough?", RuntimeWarning)

    # Check invalid parameter values

    all_ok, bad_keys = params.check_all_params()
    if exit_info is None and not all_ok:
        exit_info = ExitInformation(EXIT_INPUT_ERROR, "Bad parameters: %s" % str(bad_keys))

    # If we had an input error, quit gracefully
    if exit_info is not None:
        exit_flag = exit_info.flag
        exit_msg = exit_info.message(with_stem=True)
        results = OptimResults(None, None, None, None, 0, 0, exit_flag, exit_msg)
        return results

    # Enforce lower & upper bounds on x0
    idx = (xl < x0) & (x0 <= xl + rhobeg)
    if np.any(idx):
        warnings.warn("x0 too close to lower bound, adjusting", RuntimeWarning)
    x0[idx] = xl[idx] + rhobeg

    idx = (x0 <= xl)
    if np.any(idx):
        warnings.warn("x0 below lower bound, adjusting", RuntimeWarning)
    x0[idx] = xl[idx]

    idx = (xu - rhobeg <= x0) & (x0 < xu)
    if np.any(idx):
        warnings.warn("x0 too close to upper bound, adjusting", RuntimeWarning)
    x0[idx] = xu[idx] - rhobeg

    idx = (x0 >= xu)
    if np.any(idx):
        warnings.warn("x0 above upper bound, adjusting", RuntimeWarning)
    x0[idx] = xu[idx]

    # Call main solver (first time)
    nruns = 0
    nf = 0
    xmin, fmin, gradmin, hessmin, nf, nruns, exit_info, diagnostic_info = \
        solve_main(objfun, x0, args, rhobeg, rhoend, maxfun, nruns, nf, params, scaling_changes,
                   fixed_block_size, hess_npt)

    # Process final return values & package up
    exit_flag = exit_info.flag
    exit_msg = exit_info.message(with_stem=True)

    # Un-scale gradient & Hessian
    if scaling_changes is not None:
        if gradmin is not None:
            gradmin = gradmin / scaling_changes[1]
        if hessmin is not None:
            hessmin = hessmin / np.outer(scaling_changes[1], scaling_changes[1])

    results = OptimResults(remove_scaling(xmin, scaling_changes), fmin, gradmin, hessmin, nf, nruns, exit_flag,
                           exit_msg)

    if params("logging.save_diagnostic_info") and diagnostic_info is not None:
        df = diagnostic_info.to_dataframe()
        results.diagnostic_info = df

    return results



"""
Model
====

Maintain a class which represents an interpolating set.
This class should calculate the various geometric quantities of interest to us.


This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.

The development of this software was sponsored by NAG Ltd. (http://www.nag.co.uk)
and the EPSRC Centre For Doctoral Training in Industrially Focused Mathematical
Modelling (EP/L015803/1) at the University of Oxford. Please contact NAG for
alternative licensing.

"""

# Ensure compatibility with Python 2
from __future__ import absolute_import, division, print_function, unicode_literals

from math import sqrt
import numpy as np
import scipy.linalg as linalg

from .exit_information import *
# from .hessian import Hessian, to_upper_triangular_vector
from .trust_region import trsbox
from .util import random_orthog_directions_new_subspace_within_bounds, eval_objective, remove_scaling, model_value

__all__ = ['InterpSet', 'project_to_full_space', 'project_to_reduced_space']


class InterpSet(object):
    def __init__(self, subspace_dim, max_hess_npt, x0, f0, n=None, abs_tol=-1e20, precondition=True):
        if n is None:
            n = len(x0)
        assert 1 <= subspace_dim <= n, "Subspace dimension must be in [1..n], got %g" % subspace_dim
        assert 0 <= max_hess_npt <= n*(n+1)//2, "Max number of pts for Hessian must be in [0..n(n+1)/2], got %g" % max_hess_npt
        assert x0.shape == (n,), "x0 has wrong shape (got %s, expect (%g,))" % (str(x0.shape), n)
        # assert xl.shape == (n,), "xl has wrong shape (got %s, expect (%g,))" % (str(xl.shape), n)
        # assert xu.shape == (n,), "xu has wrong shape (got %s, expect (%g,))" % (str(xu.shape), n)
        self.n = n
        self.p = subspace_dim
        self.max_hess_npt = max_hess_npt
        self.have_hess = (self.max_hess_npt > 0)
        self.num_hess_pts = 0  # how many points have we got in the Hessian component so far?

        # Bounds (not used right now)
        self.xl = np.full((self.n,), -1e20, dtype=float)
        self.xu = np.full((self.n,), 1e20, dtype=float)

        # Interpolation points
        # All points are in the form (self.xbase + direction)
        self.points = np.full((self.p+1, self.n), np.nan, dtype=float)  # main set of interpolation directions (for gradient)
        self.points2 = None  # when add points, will be secondary set of Hessian interp points (size self.num_hess_pts, self.n)
        self.points[0,:] = x0

        # Function values
        self.fvals = np.full((self.p+1,), np.inf, dtype=float)  # objective value for each row of self.points
        self.fvals2 = None  # objective value for each row of self.points2
        self.fvals[0] = f0

        self.kopt = 0

        # Termination criteria
        self.abs_tol = abs_tol

        # Saved point - always check this value before quitting solver
        self.xsave = x0.copy()
        self.fsave = f0

        # Factorisation information
        self.simplex_factorisation_current = False
        self.Q = None  # basis for subspace, size n*p
        self.R = None  # self.points - self.xopt() directions come from Q*R, size p*p
        self.R2 = None  # self.points2 - self.xopt() directions come from Q*R2, size p*num_hess_pts
        self.precondition = precondition  # should the interpolation matrix be preconditioned?

    def xopt(self):
        return self.xpt(self.kopt)

    def fopt(self):
        return self.fval(self.kopt)

    def xpt(self, k):
        assert 0 <= k <= self.p + self.num_hess_pts, "Invalid index %g" % k
        if k <= self.p:
            return self.points[k, :]
        else:
            return self.points2[k - self.p - 1, :]

    def fval(self, k):
        assert 0 <= k <= self.p + self.num_hess_pts, "Invalid index %g" % k
        if k <= self.p:
            return self.fvals[k]
        else:
            return self.fvals2[k - self.p - 1]

    def min_objective_value(self):
        # Get termination criterion for f small: f <= abs_tol or f <= rel_tol * f0
        return self.abs_tol

    def save_point(self, x, f):
        if self.fsave is None or f <= self.fsave:
            self.xsave = x.copy()
            self.fsave = f
            return True
        else:
            return False  # this value is worse than what we have already - didn't save

    def get_final_results(self):
        # Return x and fval for optimal point (either from xsave+fsave or kopt)
        if self.fsave is None or self.fopt() <= self.fsave:  # optimal has changed since xsave+fsave were last set
            return self.xopt(), self.fopt()
        else:
            return self.xsave, self.fsave

    def initialise_interp_simplex(self, delta, objfun, args, scaling_changes, nf, maxfun, use_coord_directions=False,
                                  box_bound_thresh=0.01, full_x_thresh=6, check_for_overflow=True):
        assert delta > 0.0, "delta must be strictly positive"

        # Called upon initialisation only
        x0 = self.xpt(0)

        # Get directions
        dirns = np.zeros((self.p, self.n))  # each row is an offset from x0
        if use_coord_directions:
            idx_choices = np.random.choice(self.n, size=self.p, replace=False)
            for i in range(self.p):
                idx = idx_choices[i]
                # Decide whether to do +delta or -delta (usually via a coin toss)
                upper_gap = self.xu[idx] - x0[idx]
                lower_gap = x0[idx] - self.xl[idx]
                if min(lower_gap, upper_gap) <= box_bound_thresh * delta:
                    # If very close to boundary on at least one side, just go in the larger direction we can step
                    step = -min(delta, lower_gap) if lower_gap > upper_gap else min(delta, upper_gap)
                else:
                    step = min(delta, upper_gap) if np.random.random() >= 0.5 else -min(delta, lower_gap)

                dirns[i, idx] = step
        else:
            dirns = random_orthog_directions_new_subspace_within_bounds(self.p, delta, self.xl - x0, self.xu - x0,
                                                                        Q=None,
                                                                        box_bound_thresh=box_bound_thresh)

        # Evaluate objective at these points
        exit_info = None
        for i in range(self.p):
            x = x0 + dirns[i, :]  # point to evaluate

            if nf >= maxfun:
                exit_info = ExitInformation(EXIT_MAXFUN_WARNING, "Objective has been called MAXFUN times")
                break  # quit

            nf += 1
            f = eval_objective(objfun, remove_scaling(x, scaling_changes), args=args, eval_num=nf,
                               full_x_thresh=full_x_thresh, check_for_overflow=check_for_overflow)

            if f < self.min_objective_value():
                self.save_point(x, f)
                exit_info = ExitInformation(EXIT_SUCCESS, "Objective is sufficiently small")
                break  # quit

            self.points[i + 1, :] = x
            self.fvals[i + 1] = f

        # Choose kopt as best value so far
        self.kopt = np.argmin(self.fvals)

        self.simplex_factorisation_current = False
        return exit_info, nf

    def clear_hessian_points(self):
        # Remove all Hessian points
        if self.have_hess:
            self.points2 = None
            self.fvals2 = None
            self.num_hess_pts = 0
        return

    def append_simplex_point(self, x, f):
        assert self.p < self.n, "Cannot append points to full-dimensional interpolation set"
        self.points = np.append(self.points, x.reshape((1, self.n)), axis=0)  # append row to xpt
        self.fvals = np.append(self.fvals, f)  # append entry to fvals
        self.p += 1

        if f < self.fopt():
            self.kopt = self.p

        self.simplex_factorisation_current = False
        return

    def append_hessian_point(self, x, f):
        assert self.have_hess, "Cannot append Hessian point when using linear models"

        if self.num_hess_pts >= self.max_hess_npt:
            # Need to drop an existing point
            k_to_drop = 0  # drop oldest Hessian point
            self.points2 = np.delete(self.points2, k_to_drop, axis=0)
            self.fvals2 = np.delete(self.fvals2, k_to_drop)
            self.num_hess_pts -= 1

        if self.num_hess_pts == 0:
            self.points2 = x.reshape((1,self.n)).copy()
            self.fvals2 = np.array([f])
        else:
            self.points2 = np.append(self.points2, x.reshape((1,self.n)), axis=0)  # append row to end
            self.fvals2 = np.append(self.fvals2, f)
        self.num_hess_pts += 1
        return

    def change_simplex_point(self, k, x, f, check_not_kopt=True, move_to_hessian=True):
        # Update point k to x (w.r.t. xbase), with residual values fvec
        # If move_to_hessian, put point in self.points2
        assert 0 <= k <= self.p, "Invalid index %g" % k
        if check_not_kopt:
            assert k != self.kopt, "Cannot remove current iterate from interpolation set"
        else:
            self.save_point(self.xopt(), self.fopt())  # if we are going to delete kopt, make sure we save it first

        if move_to_hessian and self.have_hess:
            self.append_hessian_point(self.points[k, :], self.fvals[k])

        self.points[k, :] = x
        self.fvals[k] = f
        self.simplex_factorisation_current = False

        self.kopt = np.argmin(self.fvals)
        return

    def remove_simplex_point(self, k, check_not_kopt=True, move_to_hessian=True):
        assert 0 <= k <= self.p, "Invalid index %g" % k
        assert self.p >= 1, "Need to keep at least one point (iterate) in interpolation set"
        if check_not_kopt:
            assert k != self.kopt, "Cannot remove current iterate from interpolation set"

        if move_to_hessian and self.have_hess:
            self.append_hessian_point(self.points[k, :], self.fvals[k])

        self.points = np.delete(self.points, k, axis=0)  # delete row
        self.fvals = np.delete(self.fvals, k)
        self.p -= 1

        # Make sure kopt is always the best value we have in the simplex
        # Need to update even if k!=kopt (e.g. deleted point before kopt)
        self.kopt = np.argmin(self.fvals)
        self.simplex_factorisation_current = False
        return

    def remove_hessian_points(self, npts_to_drop):
        actual_npts_to_drop = max(min(npts_to_drop, self.num_hess_pts), 0)
        if actual_npts_to_drop > 0:
            # Drop oldest points (the top rows of self.points2)
            self.points2 = np.delete(self.points2, np.arange(actual_npts_to_drop), axis=0)
            self.fvals2 = np.delete(self.fvals2, np.arange(actual_npts_to_drop))
            self.num_hess_pts -= actual_npts_to_drop
        return

    def simplex_directions_from_xopt(self):
        dirns = self.points - self.xopt()  # subtract from each row of matrix (numpy does automatically)
        return np.delete(dirns, self.kopt, axis=0)  # drop kopt-th entry / kopt-th row

    def simplex_distances_to_xopt(self, include_kopt=True):
        dirns = self.points - self.xopt()  # subtract from each row of matrix (numpy does automatically)
        distances = np.linalg.norm(dirns, axis=1)  # norm of each row
        if include_kopt:
            return distances
        else:
            return np.delete(distances, self.kopt)  # drop kopt-th entry

    def factorise_simplex_system(self):
        if not self.simplex_factorisation_current:
            if self.p > 0:
                dirns = self.simplex_directions_from_xopt()  # size (p, n)
                self.Q, self.R = linalg.qr(dirns.T, mode='economic', pivoting=False)  # Q is n*p, R is p*p
            else:
                self.Q, self.R = None, None
            self.simplex_factorisation_current = True
        return

    def construct_generic_simplex_model(self, vals_to_interpolate, gradient_in_full_space=False):
        self.factorise_simplex_system()  # ensure factorisation up-to-date
        c = vals_to_interpolate[self.kopt]
        rhs = np.delete(vals_to_interpolate - c, self.kopt)  # drop kopt-th entry
        g = linalg.solve_triangular(self.R, rhs, trans='T')  # R.T \ rhs -> gradient in reduced space

        # model based at xopt
        if gradient_in_full_space:
            return c, self.project_to_full_space(g)
        else:
            return c, g

    def build_simplex_interp_model(self, gradient_in_full_space=False):
        try:
            c, g = self.construct_generic_simplex_model(self.fvals, gradient_in_full_space=gradient_in_full_space)
        except:
            return False, None, None  # flag error

        if not vector_ok(c) and vector_ok(g):
            return False, None, None  # flag error

        return True, c, g  # model based at xopt

    def simplex_lagrange_poly(self, k, gradient_in_full_space=False):
        assert 0 <= k <= self.p, "Invalid index %g" % k
        vals_to_interpolate = np.zeros((self.p + 1,))
        vals_to_interpolate[k] = 1.0  # rest are zero
        try:
            c, g = self.construct_generic_simplex_model(vals_to_interpolate, gradient_in_full_space=gradient_in_full_space)
        except:
            return False, None, None  # flag error

        if not (vector_ok(c) and vector_ok(g)):
            return False, None, None  # flag error

        return True, c, g  # model based at xopt

    def poisedness_of_each_simplex_point(self, delta=None, d=None, d_in_full_space=False):
        # Return the poisedness of each point in the interpolation set
        # if delta is set, then calculate the maximum of |L(xopt+s)| over all ||s||<=delta for each point
        # if d is set, calculate |L(xopt+d)| for each point
        assert (delta is not None or d is not None) and (delta is None or d is None), "Must specify exactly one of delta and d"
        poisedness = np.zeros((self.p + 1,))
        for k in range(self.p + 1):
            interp_ok, c, g = self.simplex_lagrange_poly(k, gradient_in_full_space=d_in_full_space)
            if not interp_ok:
                poisedness[k] = np.nan
            elif delta is not None:
                normg = np.linalg.norm(g)
                # Maximiser/minimiser of (linear) Lagrange poly in B(xopt, delta) is s = +/- delta/||g|| * g
                # Value is c +/- delta*||g||
                poisedness[k] = max(abs(c + delta * normg), abs(c - delta * normg))
            else:
                poisedness[k] = abs(c + np.dot(g, d))
        return poisedness

    def simplex_poisedness(self, delta):
        # Calculate poisedness constant in ball around xopt (everything in reduced space)
        return np.max(self.poisedness_of_each_simplex_point(delta=delta))

    def construct_generic_quadratic_model(self, simplex_vals_to_interpolate, hessian_vals_to_interpolate, Hprev=None,
                                          model_in_full_space=False):
        self.factorise_simplex_system()  # ensure factorisation up-to-date

        if self.have_hess and self.num_hess_pts > 0:
            # Build full interpolation model with c, g, and H all together
            # If use simplex to get c and g, then hessian points to only build H, then will lose interpolation at
            # simplex points

            # Merge both sets of points
            simplex_dirns_from_xopt = self.points - self.xopt()  # subtract from each row of matrix (numpy does automatically)
            simplex_dirns_from_xopt = np.delete(simplex_dirns_from_xopt, self.kopt, axis=0)  # kopt-th row
            hess_dirns_from_xopt = self.points2[:self.num_hess_pts, :] - self.xopt()
            Y_full = np.concatenate((simplex_dirns_from_xopt, hess_dirns_from_xopt), axis=0).T  # yt-xk on each column
            vals_to_interpolate = np.concatenate((simplex_vals_to_interpolate, hessian_vals_to_interpolate[:self.num_hess_pts]))
            npt = len(vals_to_interpolate)

            # Build interpolation matrix (in reduced space)
            Y_red = self.Q.T.dot(Y_full)  # (yt-xk) in reduced coords on each column
            # Note: rows of Y_red for Hessian points are approx (yt-xk not in col Q)
            distances_to_xopt = np.linalg.norm(Y_full, axis=1)
            approx_delta = sqrt(np.max(distances_to_xopt)) if self.precondition else 1.0  # largest distance to xopt ~ delta
            A, left_scaling, right_scaling = build_interpolation_matrix(Y_red, approx_delta=approx_delta)

            # Solve linear system
            c = vals_to_interpolate[self.kopt]
            if npt == self.p + 1 or npt == (self.p + 1) * (self.p + 2) // 2:
                rhs = np.delete(vals_to_interpolate - c, self.kopt)  # drop kopt-th entry
            else:
                rhs = np.zeros((npt + self.p - 1,))
                rhs[:npt - 1] = np.delete(vals_to_interpolate - c, self.kopt)
                if Hprev is not None:
                    # Modified to be minimum *change* in Hessian, rather than minimum Hessian
                    for t in range(npt - 1):
                        dx = Y_red[:, t]
                        rhs[t] = rhs[t] - 0.5 * np.dot(dx, Hprev.dot(dx))  # include old Hessian
            coeffs = np.linalg.solve(A, rhs * left_scaling) * right_scaling  # solve linear system

            # Construct model (note: c defined above)
            if npt == self.p + 1:
                g = coeffs.copy()
                H = None
            elif npt == (self.p + 1) * (self.p + 2) // 2:
                g = coeffs[:self.p]
                # H = Hessian(self.p, coeffs[self.p:])  # rest of coeffs are upper triangular part of Hess
                H = build_symmetric_matrix_from_vector(self.p, coeffs[self.p:])
            else:
                g = coeffs[npt - 1:]  # last n values
                H = np.zeros((self.p, self.p)) if Hprev is None else Hprev  # min change Hessian?
                for i in range(npt - 1):
                    dx = Y_red[:, i]
                    H += coeffs[i] * np.outer(dx, dx)

        else:
            # Only building linear models - easy case
            c, g = self.construct_generic_simplex_model(simplex_vals_to_interpolate, gradient_in_full_space=False)
            H = None

        if model_in_full_space:
            cfull = c
            gfull = self.project_to_full_space(g)
            Hfull = np.dot(self.Q, np.dot(H, self.Q.T)) if H is not None else None
            return cfull, gfull, Hfull  # based at self.xopt()
        else:
            return c, g, H  # based at self.xopt()

    def build_quadratic_interp_model(self, Hprev=None, model_in_full_space=False):
        # Note: Hprev needs to be p*p (in *current* self.Q basis)
        try:
            c, g, H = self.construct_generic_quadratic_model(self.fvals, self.fvals2, Hprev=Hprev,
                                                             model_in_full_space=model_in_full_space)
        except:
            return False, None, None, None  # flag error

        if self.have_hess and self.num_hess_pts > 0:
            if not (vector_ok(c) and vector_ok(g) and vector_ok(H)):
                return False, None, None, None  # flag error
        else:  # don't check Hessian
            if not (vector_ok(c) and vector_ok(g) and H is None):
                return False, None, None, None  # flag error

        return True, c, g, H  # model based at xopt

    def lagrange_poly(self, k, model_in_full_space=False):
        assert 0 <= k < self.p + 1 + self.num_hess_pts, "Invalid k (got %g, must be in [0,%g])" % (k, self.p+1+self.num_hess_pts)
        simplex_vals_to_interpolate = np.zeros((self.p+1,))
        hessian_vals_to_interpolate = np.zeros((self.num_hess_pts))
        if k < self.p+1:
            simplex_vals_to_interpolate[k] = 1.0
        else:
            hessian_vals_to_interpolate[k-self.p-1] = 1.0
        return self.construct_generic_quadratic_model(simplex_vals_to_interpolate, hessian_vals_to_interpolate,
                                                      Hprev=None, model_in_full_space=model_in_full_space)

    def poisedness_in_reduced_space(self, delta):
        # Calculate poisedness constant in ball around xopt (everything in reduced space)
        lmax = None
        for k in range(self.p + 1 + self.num_hess_pts):
            c, g, H = self.lagrange_poly(k, model_in_full_space=False)
            if H is not None:
                # Quadratic models
                dmin, _, _ = trsbox(self.project_to_reduced_space(self.xopt()), g, H, -1e20*np.ones((self.p,)), 1e20*np.ones((self.p,)), delta)
                dmax, _, _ = trsbox(self.project_to_reduced_space(self.xopt()), -g, -H, -1e20 * np.ones((self.p,)), 1e20 * np.ones((self.p,)), delta)
                l = max(abs(c + model_value(g, H, dmin)), abs(c + model_value(g, H, dmax)))
            else:
                # Linear models
                dmin = -delta * g / np.linalg.norm(g)
                dmax = delta * g / np.linalg.norm(g)
                l = max(abs(c + np.dot(g, dmin)), abs(c + np.dot(g, dmax)))
            if lmax is None or l > lmax:
                lmax = l
        return lmax

    def project_to_full_space(self, x):
        assert self.simplex_factorisation_current, "Cannot project, factorisation is invalid"
        return project_to_full_space(self.Q, x)

    def project_to_reduced_space(self, x):
        assert self.simplex_factorisation_current, "Cannot project, factorisation is invalid"
        return project_to_reduced_space(self.Q, x)


def project_to_full_space(Q, x):
    return Q.dot(x)


def project_to_reduced_space(Q, x):
    return Q.T.dot(x)


def vector_ok(x):
    return np.all(np.isfinite(x))


def build_interpolation_matrix(Y, approx_delta=1.0):
    # Routine from Py-BOBYQA
    # Y has columns Y[:,t] = yt - xk
    n, p = Y.shape  # p = npt-1
    assert n + 1 <= p + 1 <= (n + 1) * (n + 2) // 2, "npt must be in range [n+1, (n+1)(n+2)/2]"

    # What scaling was applied to each part of the matrix?
    # A(scaled) = diag(left_scaling) * A(unscaled) * diag(right_scaling)

    if p == n:  # linear models
        A = Y.T / approx_delta
        left_scaling = np.ones((n,))  # no left scaling
        right_scaling = np.ones((n,)) / approx_delta
    elif p + 1 == (n + 1) * (n + 2) // 2:  # fully quadratic models
        A = np.zeros((p, p))
        A[:, :n] = Y.T / approx_delta
        for i in range(p):
            A[i, n:] = get_entry_vector_from_symmetric_matrix(np.outer(Y[:, i], Y[:, i]) - 0.5 * np.diag(np.square(Y[:, i]))) / (approx_delta ** 2)
            # A[i, n:] = to_upper_triangular_vector(np.outer(Y[:, i], Y[:, i]) - 0.5 * np.diag(np.square(Y[:, i]))) / (approx_delta ** 2)
        left_scaling = np.ones((p,))  # no left scaling
        right_scaling = np.ones((p,))
        right_scaling[:n] = 1.0 / approx_delta
        right_scaling[n:] = 1.0 / (approx_delta ** 2)
    else:  # underdetermined quadratic models
        A = np.zeros((p + n, p + n))
        for i in range(p):
            for j in range(p):
                A[i, j] = 0.5 * np.dot(Y[:, i], Y[:, j]) ** 2 / (approx_delta ** 4)
        A[:p, p:] = Y.T / approx_delta
        A[p:, :p] = Y / approx_delta
        left_scaling = np.ones((p + n,))
        right_scaling = np.ones((p + n,))
        left_scaling[:p] = 1.0 / (approx_delta ** 2)
        left_scaling[p:] = approx_delta
        right_scaling[:p] = 1.0 / (approx_delta ** 2)
        right_scaling[p:] = approx_delta
    return A, left_scaling, right_scaling


def build_symmetric_matrix_from_vector(n, entries):
    assert entries.shape == (n*(n+1)//2,), "Entries vector has wrong size, got %g, expect %g (for n=%g)" % (len(entries), n*(n+1)//2, n)
    A = np.zeros((n, n))
    ih = -1
    for j in range(n):  # j = 0, ..., n-1
        for i in range(j + 1):  # i = 0, ..., j
            ih += 1
            A[i, j] = entries[ih]
            A[j, i] = entries[ih]
    return A


def get_entry_vector_from_symmetric_matrix(A):
    n = A.shape[0]
    hq = np.zeros((n*(n+1)//2,))
    ih = -1
    for j in range(n):  # j = 0, ..., n-1
        for i in range(j + 1):  # i = 0, ..., j
            ih += 1
            hq[ih] = A[i,j]
    return hq


"""
Parameters
====

A container class for all the solver parameter values.


This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.

The development of this software was sponsored by NAG Ltd. (http://www.nag.co.uk)
and the EPSRC Centre For Doctoral Training in Industrially Focused Mathematical
Modelling (EP/L015803/1) at the University of Oxford. Please contact NAG for
alternative licensing.

"""

# Ensure compatibility with Python 2
from __future__ import absolute_import, division, print_function, unicode_literals

__all__ = ['ParameterList']


class ParameterList(object):
    def __init__(self, n, maxfun, objfun_has_noise=False):
        self.n = n

        self.params = {}
        self.params["general.safety_step_thresh"] = 0.5  # safety step called if ||d|| <= thresh * rho
        self.params["general.check_objfun_for_overflow"] = True
        # Logging
        self.params["logging.n_to_print_whole_x_vector"] = 6
        self.params["logging.save_diagnostic_info"] = False
        self.params["logging.save_poisedness"] = False
        self.params["logging.save_xk"] = False
        # Trust Region Radius management
        self.params["tr_radius.eta1"] = 0.1
        self.params["tr_radius.eta2"] = 0.7
        self.params["tr_radius.gamma_dec"] = 0.98 if objfun_has_noise else 0.5
        self.params["tr_radius.gamma_inc"] = 2.0
        self.params["tr_radius.gamma_inc_overline"] = 4.0
        self.params["tr_radius.delta_max"] = 1.0e10
        self.params["tr_radius.update_method"] = 'bobyqa'
        self.params["tr_radius.alpha1"] = 0.9 if objfun_has_noise else 0.1
        self.params["tr_radius.alpha2"] = 0.95 if objfun_has_noise else 0.5
        self.params["tr_radius.use_rho"] = False  # use delta/rho combination?
        # Least-Squares objective threshold
        self.params["model.abs_tol"] = -1e20
        self.params["model.precondition_linear_system"] = True
        # Slow progress thresholds
        self.params["slow.history_for_slow"] = 5
        self.params["slow.thresh_for_slow"] = 1e-4
        self.params["slow.max_slow_iters"] = 20 * n
        # Geometry Management
        self.params["geometry.use_coord_directions"] = False  # make a block coordinate method?
        self.params["geometry.direcion_box_bound_thresh"] = 0.01  # how close to bounds before random sampling is changed
        self.params["geometry.safety_steps_before_redraw"] = 1  # how many consecutive safety steps before redrawing a model (1 = every time)
        # Adaptive block
        self.params["adaptive.simplex_save_frac_successful"] = 1.0  # proportion of simplex points to save (successful steps)
        self.params["adaptive.simplex_save_frac_unsuccessful"] = 0.9  # proportion of simplex points to save (unsuccessful steps)
        self.params["adaptive.min_change_hess"] = True  # interp using minimum Hessian change
        # Hessian
        self.params["hessian.forget_on_unsuccessful"] = False  # drop saved Hessian on unsuccessful iterations?
        self.params["hessian.drop_interp_points_on_unsuccessful"] = False  # clear_hessian_points on unsuccessful?
        self.params_changed = {}
        for p in self.params:
            self.params_changed[p] = False

    def __call__(self, key, new_value=None):  # self(key) or self(key, new_value)
        if key in self.params:
            if new_value is None:
                return self.params[key]
            else:
                if self.params_changed[key]:
                    raise ValueError("Trying to update parameter '%s' for a second time" % key)
                self.params[key] = new_value
                self.params_changed[key] = True
                return self.params[key]
        else:
            raise ValueError("Unknown parameter '%s'" % key)

    def param_type(self, key):
        # Use the check_* methods below, but switch based on key
        if key == "general.safety_step_thresh":
            type_str, nonetype_ok, lower, upper = 'float', False, 0.0, None
        elif key == "general.check_objfun_for_overflow":
            type_str, nonetype_ok, lower, upper = 'bool', False, None, None
        elif key == "logging.n_to_print_whole_x_vector":
            type_str, nonetype_ok, lower, upper = 'int', False, 0, None
        elif key == "logging.save_diagnostic_info":
            type_str, nonetype_ok, lower, upper = 'bool', False, None, None
        elif key == "logging.save_poisedness":
            type_str, nonetype_ok, lower, upper = 'bool', False, None, None
        elif key == "logging.save_xk":
            type_str, nonetype_ok, lower, upper = 'bool', False, None, None
        elif key == "tr_radius.eta1":
            type_str, nonetype_ok, lower, upper = 'float', False, 0.0, 1.0
        elif key == "tr_radius.eta2":
            type_str, nonetype_ok, lower, upper = 'float', False, 0.0, 1.0
        elif key == "tr_radius.gamma_dec":
            type_str, nonetype_ok, lower, upper = 'float', False, 0.0, 1.0
        elif key == "tr_radius.gamma_inc":
            type_str, nonetype_ok, lower, upper = 'float', False, 1.0, None
        elif key == "tr_radius.gamma_inc_overline":
            type_str, nonetype_ok, lower, upper = 'float', False, 1.0, None
        elif key == "tr_radius.delta_max":
            type_str, nonetype_ok, lower, upper = 'float', False, 0.0, None
        elif key == "tr_radius.update_method":
            type_str, nonetype_ok, lower, upper = 'str', False, None, None
        elif key == "tr_radius.alpha1":
            type_str, nonetype_ok, lower, upper = 'float', False, 0.0, 1.0
        elif key == "tr_radius.alpha2":
            type_str, nonetype_ok, lower, upper = 'float', False, 0.0, 1.0
        elif key == "tr_radius.use_rho":
            type_str, nonetype_ok, lower, upper = 'bool', False, None, None
        elif key == "model.abs_tol":
            type_str, nonetype_ok, lower, upper = 'float', False, None, None
        elif key == "model.precondition_linear_system":
            type_str, nonetype_ok, lower, upper = 'bool', False, None, None
        elif key == "slow.history_for_slow":
            type_str, nonetype_ok, lower, upper = 'int', False, 0, None
        elif key == "slow.thresh_for_slow":
            type_str, nonetype_ok, lower, upper = 'float', False, 0, None
        elif key == "slow.max_slow_iters":
            type_str, nonetype_ok, lower, upper = 'int', False, 0, None
        elif key == "geometry.use_fin_diff":
            type_str, nonetype_ok, lower, upper = 'bool', False, None, None
        elif key == "geometry.use_coord_directions":
            type_str, nonetype_ok, lower, upper = 'bool', False, None, None
        elif key == "geometry.direcion_box_bound_thresh":
            type_str, nonetype_ok, lower, upper = 'float', False, 0.0, 1.0
        elif key == "geometry.safety_steps_before_redraw":
            type_str, nonetype_ok, lower, upper = 'int', False, 1, None
        elif key == "geometry.vsucc_steps_before_redraw":
            type_str, nonetype_ok, lower, upper = 'int', False, 1, None
        # elif key == "geometry.adaptive_dimension":
        #     type_str, nonetype_ok, lower, upper = 'bool', False, None, None
        elif key == "adaptive.simplex_save_frac_successful":
            type_str, nonetype_ok, lower, upper = 'float', False, 0.0, 1.0
        elif key == "adaptive.simplex_save_frac_unsuccessful":
            type_str, nonetype_ok, lower, upper = 'float', False, 0.0, 1.0
        elif key == "adaptive.min_change_hess":
            type_str, nonetype_ok, lower, upper = 'bool', False, None, None
        elif key == "hessian.forget_on_unsuccessful":
            type_str, nonetype_ok, lower, upper = 'bool', False, None, None
        elif key == "hessian.drop_interp_points_on_unsuccessful":
            type_str, nonetype_ok, lower, upper = 'bool', False, None, None
        else:
            assert False, "ParameterList.param_type() has unknown key: %s" % key
        return type_str, nonetype_ok, lower, upper

    def check_param(self, key, value):
        type_str, nonetype_ok, lower, upper = self.param_type(key)
        if type_str == 'int':
            return check_integer(value, lower=lower, upper=upper, allow_nonetype=nonetype_ok)
        elif type_str == 'float':
            return check_float(value, lower=lower, upper=upper, allow_nonetype=nonetype_ok)
        elif type_str == 'bool':
            return check_bool(value, allow_nonetype=nonetype_ok)
        elif type_str == 'str':
            return check_str(value, allow_nonetype=nonetype_ok)
        else:
            assert False, "Unknown type_str '%s' for parameter '%s'" % (type_str, key)

    def check_all_params(self):
        bad_keys = []
        for key in self.params:
            if not self.check_param(key, self.params[key]):
                bad_keys.append(key)
        return len(bad_keys) == 0, bad_keys


def check_integer(val, lower=None, upper=None, allow_nonetype=False):
    # Check that val is an integer and (optionally) that lower <= val <= upper
    if val is None:
        return allow_nonetype
    elif not isinstance(val, int):
        return False
    else:  # is integer
        return (lower is None or val >= lower) and (upper is None or val <= upper)


def check_float(val, lower=None, upper=None, allow_nonetype=False):
    # Check that val is a float and (optionally) that lower <= val <= upper
    if val is None:
        return allow_nonetype
    elif not isinstance(val, float):
        return False
    else:  # is integer
        return (lower is None or val >= lower) and (upper is None or val <= upper)


def check_bool(val, allow_nonetype=False):
    if val is None:
        return allow_nonetype
    else:
        return isinstance(val, bool)


def check_str(val, allow_nonetype=False):
    if val is None:
        return allow_nonetype
    else:
        return isinstance(val,str) or isinstance(val, unicode)

